//
//  ChannelListTableViewController.swift
//  SampleApp
//
//  Created by Federico Zanetello on 10/4/18.
//  Copyright © 2018 David Zhang. All rights reserved.
//

import UIKit
import SwiftUI
import AmitySDK
import Combine

/*
 * Note:
 *
 * Look into `ChannelListDataSource` class to see the sdk implementation to fetch channel
 * list.
 */
final class ChannelListTableViewController: UIViewController, DataSourceListener, UITableViewDataSource, UITableViewDelegate {
    
    @IBOutlet weak var filterButton: UIButton!
    @IBOutlet weak var tableView: UITableView!
    @IBOutlet weak var updateButton: UIButton!
    @IBOutlet weak var singleChannelButton: UIButton!
    
    var presenceRepo = AmityChannelPresenceRepository(client: AmityManager.shared.client!)
    
    // To be injected.
    var repository: AmityChannelRepository! {
        didSet {
            dataSource = ChannelListDataSource(channelRepository: repository)
            dataSource?.dataSourceObserver = self
        }
    }
    
    weak var delegate: ChannelsTableViewControllerDelegate?
    public var channelType: AmityChannelType = .standard
    
    private var dataSource: ChannelListDataSource?
    private lazy var cellConfigurator = ChannelListTableViewCellConfigurator()
    
    let refreshControl = UIRefreshControl()
    
    var cancellable: AnyCancellable?
    var presencesDict = [String: AmityChannelPresence]()
    
    // MARK: View Lifecycle
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        presenceRepo.unsyncAllChannelPresence()
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        let nib = UINib(nibName: ChannelListTableViewCell.identifier, bundle: nil)
        tableView.register(nib, forCellReuseIdentifier: ChannelListTableViewCell.identifier)
        //let nib = UINib(nibName: SampleAppTableViewCell.identifier, bundle: nil)
        //tableView.register(nib, forCellReuseIdentifier: SampleAppTableViewCell.identifier)
        
        refreshControl.addTarget(self, action: #selector(fetchChannelList), for: .valueChanged)
        
        tableView.dataSource = self
        tableView.delegate = self
        tableView.refreshControl = refreshControl
        
        filterButton.addTarget(self, action: #selector(onFilterButtonTap), for: .touchUpInside)
        updateButton.addTarget(self, action: #selector(onUpdateButtonTap), for: .touchUpInside)
        singleChannelButton.addTarget(self, action: #selector(onSingleChannelButtonTap), for: .touchUpInside)
        
        cancellable = presenceRepo.getSyncingChannelPresence().sink(receiveCompletion: { completion in
            
        }, receiveValue: { [weak self] channelPresences in
            guard let self = self else { return }
            
            self.renderChannelPresences(channelPresences: channelPresences)
        })
    }
    
    func renderChannelPresences(channelPresences: [AmityChannelPresence]) {
        var shouldUpdatePresences = false
        
        if channelPresences.isEmpty {
            shouldUpdatePresences = !presencesDict.isEmpty // If it was not empty before, update tableview
            presencesDict = [:]
        } else {
            channelPresences.forEach { self.presencesDict[$0.channelId] = $0 }
            shouldUpdatePresences = true
        }
        
        guard shouldUpdatePresences else { return }
        
        tableView.reloadData()
    }
    
    @objc func onFilterButtonTap() {
        let filterView = ChannelFilterView {
            self.fetchChannelList()
        }
        let hostingController = UIHostingController(rootView: filterView)
        self.navigationController?.pushViewController(hostingController, animated: true)
    }
    
    @objc func onUpdateButtonTap() {
        let hostingController = UIHostingController(rootView: ChannelUpdateView())
        self.navigationController?.pushViewController(hostingController, animated: true)
    }
    
    @objc func onSingleChannelButtonTap() {
        let hostingController = UIHostingController(rootView: SingleChannelQueryView())
        self.navigationController?.pushViewController(hostingController, animated: true)
    }
    
    /// Fires a new request to fetch all the channels matching the current filters and tags preferences.
    @objc func fetchChannelList() {
        guard let dataSource = dataSource else { return }
        let channelFilter = UserDefaults.standard.filter
        let includingTags = UserDefaults.standard.includingTags
        let excludingTags = UserDefaults.standard.excludingTags
        channelType = UserDefaults.standard.channelTypeFilter

        Log.add(info: "Channel Type: \(UserDefaults.standard.channelTypeFilter.description)")
        Log.add(info: "Filter: \(channelFilter.description)")
        Log.add(info: "Including Tags: \(includingTags), Excluding Tags: \(excludingTags)")
        
        refreshControl.endRefreshing()

        if (channelFilter == .userIsNotMember || channelFilter == .all) && channelType == .private {
            dataSource.fetchChannels(
                channelType: .unknown,
                channelFilter: channelFilter,
                includingTags: includingTags,
                excludingTags: excludingTags,
                channelTypes: Set(arrayLiteral: "private")
            )
        } else {
            dataSource.fetchChannels(
                channelType: channelType,
                channelFilter: channelFilter,
                includingTags: includingTags,
                excludingTags: excludingTags
            )
        }
    }
    
    // MARK: UITableViewDataSource
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return dataSource?.numberOfChannels() ?? 0
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        
        guard let channel: AmityChannel = dataSource?.channel(for: indexPath) else { return UITableViewCell() }

        let cell: UITableViewCell = tableView.dequeueReusableCell(withIdentifier: ChannelListTableViewCell.identifier, for: indexPath)
        guard let sampleAppCell = cell as? ChannelListTableViewCell else { return UITableViewCell() }
        
        cellConfigurator.configure(sampleAppCell, with: channel)
        
        if let presenceInfo = presencesDict[channel.channelId], presenceInfo.isAnyMemberOnline {
            sampleAppCell.onlineIndicator.isHidden = false
        } else {
            sampleAppCell.onlineIndicator.isHidden = true
        }
        
        return sampleAppCell
    }
    
    func tableView(_ tableView: UITableView, willDisplay cell: UITableViewCell, forRowAt indexPath: IndexPath) {
        guard let channel: AmityChannel = dataSource?.channel(for: indexPath), channel.channelType == .conversation else { return }
        
        presenceRepo.syncChannelPresence(id: channel.channelId)
    }
    
    func tableView(_ tableView: UITableView, didEndDisplaying cell: UITableViewCell, forRowAt indexPath: IndexPath) {
        guard let channel: AmityChannel = dataSource?.channel(for: indexPath), channel.channelType == .conversation else { return }
        
        presenceRepo.unsyncChannelPresence(id: channel.channelId)
    }
    
    // MARK: UITableViewDelegate
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        guard let channel: AmityChannel = dataSource?.channel(for: indexPath) else { return }
        
        let action1 = UIAlertAction(title: "Join Channel", style: .default) { [weak self] action in
            self?.delegate?.joinChannel(channel: channel)
        }
        
        let action2 = UIAlertAction(title: "View Details", style: .default) { [weak self] action in
            let detailController = UIHostingController(rootView: ChannelDetailsView(detail: ChannelDetail(object: channel)))
            self?.navigationController?.pushViewController(detailController, animated: true)
        }
        
        let action3 = UIAlertAction(title: "Channel Real Time Events", style: .default) { [weak self] action in
            let realTimeController = ChannelRealTimeEventController()
            realTimeController.manager = ChannelRealTimeEventManager(channel: channel)
            
            // start observe only for auto subscription channel
            if channel.channelType == .community || channel.channelType == .conversation {
                realTimeController.shouldStartObserve = true
            }
            
            self?.navigationController?.pushViewController(realTimeController, animated: true)
        }
        
        let action4 = UIAlertAction(title: "Channel Mark Read", style: .default) { [weak self] action in
            Task {
                do {
                    try await channel.markAsRead()
                    guard let self else { return }
                    AppUtility.showAlert(in: self, title: "", message: "Mark read success")
                } catch let error {
                    guard let self else { return }
                    AppUtility.showAlert(in: self, title: "Error", message: "\(error.localizedDescription)")
                }
            }
        }
        
        AppUtility.showActionSheet(in: self, title: "", message: "What would you like to do?", actions: [action1, action2, action3, action4])
    }
    
    func tableView(_ tableView: UITableView, trailingSwipeActionsConfigurationForRowAt indexPath: IndexPath) -> UISwipeActionsConfiguration? {
        let channelId = dataSource?.channel(for: indexPath)?.channelId ?? ""
        let action1 = UIContextualAction(style: .normal, title: "Copy Id") { (action, view, completion) in
            UIPasteboard.general.string = channelId
            completion(true)
        }
        
        let config = UISwipeActionsConfiguration(actions: [action1])
        return config
    }
    
    // MARK: UIScrollViewDelegate
    
    func scrollViewWillEndDragging(_ scrollView: UIScrollView, withVelocity velocity: CGPoint, targetContentOffset: UnsafeMutablePointer<CGPoint>) {
        let height = scrollView.frame.size.height
        let contentYoffset = scrollView.contentOffset.y
        let distanceFromBottom = scrollView.contentSize.height - contentYoffset
        if distanceFromBottom < height {
            dataSource?.fetchMoreChannels()
        }
    }
    
    // MARK: DataSourceListener
    
    func didUpdateDataSource() {
        tableView.reloadData()
    }
}

extension AmityChannelQueryFilter: CustomStringConvertible {
    public var description: String {
        switch self {
        case .all: return "All"
        case .userIsMember: return "User Is Member"
        case .userIsNotMember: return "User Is Not A Member"
        @unknown default:
            assertionFailure("Unknown")
            return "Unknown"
        }
    }
}
