//
//  SubChannelListController.swift
//  SampleApp
//
//  Created by Amity on 28/10/2565 BE.
//  Copyright © 2565 BE David Zhang. All rights reserved.
//

import Foundation
import UIKit
import AmitySDK
import SwiftUI
import Combine

class SubChannelListController: UIViewController, UITableViewDataSource, UITableViewDelegate {
    
    let tableView = UITableView()
    let totalMessageCountLabel = UILabel()
    
    var channel: AmityChannel?
    let manager = SubChannelManager()
    
    let client = AmityManager.shared.client
    
    var channelId: String = ""
    var fetchDeletedSubChannels = false
    var excludeMainSubChannel = false
    let messageRepository = AmityMessageRepository(client: AmityManager.shared.client!)
    let channelRepository = AmityChannelRepository(client: AmityManager.shared.client!)
    var messageToken: AmityNotificationToken?
    
    var cancellable: AnyCancellable?
    
    override func loadView() {
        view = UIView()
        view.backgroundColor = .white
    }
    
    private var disposeBag: Set<AnyCancellable> = []
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        self.title = "Sub Channels"
        trackUnreadCount()
        addLeaveButton()

        totalMessageCountLabel.textAlignment = .right
        tableView.translatesAutoresizingMaskIntoConstraints = false
        totalMessageCountLabel.translatesAutoresizingMaskIntoConstraints = false
        view.addSubview(totalMessageCountLabel)
        view.addSubview(tableView)
    
        NSLayoutConstraint.activate([
            totalMessageCountLabel.leadingAnchor.constraint(equalTo: view.leadingAnchor),
            totalMessageCountLabel.trailingAnchor.constraint(equalTo: view.trailingAnchor),
            totalMessageCountLabel.topAnchor.constraint(equalTo: view.safeAreaLayoutGuide.topAnchor),
            totalMessageCountLabel.bottomAnchor.constraint(equalTo: tableView.topAnchor),
            tableView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
            tableView.trailingAnchor.constraint(equalTo: view.trailingAnchor),
            tableView.topAnchor.constraint(equalTo: totalMessageCountLabel.bottomAnchor),
            tableView.bottomAnchor.constraint(equalTo: view.bottomAnchor),
        ])
        
        let nib = UINib(nibName: SampleAppTableViewCell.identifier, bundle: nil)
        tableView.register(nib, forCellReuseIdentifier: SampleAppTableViewCell.identifier)
        tableView.dataSource = self
        tableView.delegate = self
        
        let moreOptionButton = UIBarButtonItem(image: UIImage(systemName: "ellipsis.circle"), style: .plain, target: self, action: #selector(onMoreOptionButtonTap))
        
        let filterButton = UIBarButtonItem(image: UIImage(systemName: "line.3.horizontal.decrease.circle"), style: .plain, target: self, action: #selector(onFilterButtonTap))
        self.navigationItem.rightBarButtonItems = [moreOptionButton, filterButton]
        
        guard let channelId = channel?.channelId else { return }
        self.channelId = channelId
        
        loadSubChannels()
    }
    
    func trackUnreadCount() {
        client?.getUserUnread().sink(receiveValue: { [weak self] userUnread in
            DispatchQueue.main.async {
                var textStatus = "Total Unread Count: \(userUnread.unreadCount)"
                if userUnread.hasMentioned {
                    textStatus += " @"
                }
                self?.totalMessageCountLabel.text = textStatus
            }
        }).store(in: &disposeBag)
    }
    
    func loadSubChannels() {
        manager.fetchSubChannels(channelId: channelId, isDeleted: fetchDeletedSubChannels, excludeMainChannel: excludeMainSubChannel) { [weak self] subChannels, error in
            // Reload tableview
            self?.tableView.reloadData()
        }
        
        // If there is any changes due to deletion, reload tableview again
        cancellable = nil
        cancellable = manager.$subChannels.sink { [weak self] value in
            self?.tableView.reloadData()
        }
    }
    
    @objc func onMoreOptionButtonTap() {
        let createAction = UIAlertAction(title: "Create SubChannel", style: .default) { [weak self] action in
            let hostingController = UIHostingController(rootView: CreateSubChannelView(channelId: self?.channelId ?? ""))
            hostingController.title = "Create"
            self?.navigationController?.pushViewController(hostingController, animated: true)
        }
        
        let singleQueryAction = UIAlertAction(title: "Get SubChannel", style: .default) { [weak self] action in
            let hostingController = UIHostingController(rootView: SingleSubChannelQueryView())
            self?.navigationController?.pushViewController(hostingController, animated: true)
        }

        AppUtility.showActionSheet(in: self, title: "", message: "SubChannels", actions: [createAction, singleQueryAction])
    }
    
    @objc func onAddButtonTap() {
        let hostingController = UIHostingController(rootView: CreateSubChannelView(channelId: self.channelId))
        hostingController.title = "Create"
        self.navigationController?.pushViewController(hostingController, animated: true)
    }
    
    @objc func onFilterButtonTap() {
        AppUtility.showActionSheet(in: self, title: "Filter SubChannels", message: "Filter SubChannels By:", actions: [
            UIAlertAction(title: "IsDeleted: True", style: .default) { [weak self] action in
                self?.fetchDeletedSubChannels = true
                self?.loadSubChannels()
            },
            
            UIAlertAction(title: "IsDeleted: False", style: .default) { [weak self] action in
                self?.fetchDeletedSubChannels = false
                self?.loadSubChannels()
            },
            
            UIAlertAction(title: "Exclude Main Channel", style: .default) { [weak self] action in
                self?.excludeMainSubChannel = true
                self?.loadSubChannels()
            },
            
            UIAlertAction(title: "Include Main Channel", style: .default) { [weak self] action in
                self?.excludeMainSubChannel = false
                self?.loadSubChannels()
            }
        ])
    }
    
    func subscribeToSubChanenl(subChannel: AmitySubChannel) {
        subChannel.subscribeEvent { [weak self] isSuccess, error in
            guard let strongSelf = self else { return }
            
            let message = isSuccess ? "Subscribed Successfully: \(subChannel.displayName)" : "Failed to subscribe event"
            AppUtility.showAlert(in: strongSelf, title: "", message: message, action: nil)
        }
    }
    
    func unSubscribeSubChannel(subChannel: AmitySubChannel) {
        subChannel.unSubscribeEvent { [weak self] isSuccess, error in
            guard let strongSelf = self else { return }
            
            let message = isSuccess ? "Subscribed Successfully: \(subChannel.displayName)" : "Failed to subscribe event"
            AppUtility.showAlert(in: strongSelf, title: "", message: message, action: nil)
        }
    }
    
    private func addLeaveButton() {
        let leaveButton = UIBarButtonItem(title: "Leave",
                                          style: .plain,
                                          target: self,
                                          action: #selector(leave))
        navigationItem.setLeftBarButton(leaveButton,
                                        animated: true)
        navigationItem.leftItemsSupplementBackButton = true
    }
    
    @objc private func leave() {
        Task {
            do {
                try await channelRepository.leaveChannel(channelId: channelId)
                navigationController?.popViewController(animated: true)
            } catch {
                // Even though showLeaveError is marked with @MainActor, it still does not run on main thread
                // unless explicitly called on main thread
                await MainActor.run {
                    showLeaveError(error)
                }
            }
        }
    }
    
    @MainActor
    private func showLeaveError(_ error: Error) {
        var message: String {
            if error.isAmityErrorCode(.moderatorUnableToLeaveCommunity) {
                return "You’re the only moderator in this group. To leave community, nominate other members to moderator role."
            } else {
                return error.localizedDescription
            }
        }
        
        AppUtility.showAlert(in: self, title: "Unable to leave community", message: message)
    }
}

extension SubChannelListController {
    
    func numberOfSections(in tableView: UITableView) -> Int {
        return 1
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return manager.subChannels.count
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: SampleAppTableViewCell.identifier) as? SampleAppTableViewCell
        
        let item = manager.subChannels[indexPath.row]
        cell?.titleLabel?.text = item.displayName
        
        if let previewMessage = item.preview {
            let user = previewMessage.user?.displayName ?? "Preview:"
            
            if previewMessage.dataType == .text {
                let messageData = previewMessage.data?["text"] as? String ?? ""
                cell?.subtitleLabel.text = "\(user): \(messageData)"
            } else {
                cell?.subtitleLabel.text = "\(user) sent \(previewMessage.dataType.description) message"
            }
        } else {
            cell?.subtitleLabel.text = ""
        }
        
        var detailText: String = ""
        if item.unreadCount != 0 {
            detailText = "🆕 \(item.unreadCount)"
        }
        let mentionIndicator = item.hasMentioned ? " @" : ""
        detailText += mentionIndicator
        cell?.detailLabel.text = detailText
        return cell!
    }
    
    func scrollViewWillEndDragging(_ scrollView: UIScrollView, withVelocity velocity: CGPoint, targetContentOffset: UnsafeMutablePointer<CGPoint>) {
        let frameHeight = scrollView.frame.size.height
        let contentOffset = scrollView.contentOffset.y
        let contentHeight = scrollView.contentSize.height
        
        let distanceFromBottom = contentHeight - contentOffset
        if distanceFromBottom < frameHeight {
            manager.fetchMoreSubChannels()
        }
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        let subChannel = manager.subChannels[indexPath.row]
        // If its deleted subchannel, do not allow user to do anything
        guard !subChannel.isDeleted else { return }
        
        let editAction = UIAlertAction(title: "Edit", style: .default) { [weak self] action in
            self?.showEditScreen(subChannel: subChannel)
        }
        
        let chatAction = UIAlertAction(title: "Chat Screen", style: .default) { [weak self] action in
            self?.displayChatScreen(subChannel: subChannel.object)
        }
        
        let viewDetail = UIAlertAction(title: "View Details", style: .default) { [weak self] action in
            let detailController = UIHostingController(rootView: SubChannelDetailsView(viewModel: SubChannelDetailsViewModel(subChannelId: subChannel.id)))
            
            self?.navigationController?.pushViewController(detailController, animated: true)
        }
        
        let subscribeAction = UIAlertAction(title: "Subscribe Events", style: .default) { [weak self] action in
            self?.subscribeToSubChanenl(subChannel: subChannel.object)
        }
        
        let unSubscribeAction = UIAlertAction(title: "Un Subscribe Events", style: .default) { [weak self] action in
            self?.unSubscribeSubChannel(subChannel: subChannel.object)
        }
        
        let copyAction = UIAlertAction(title: "Copy SubChannelId", style: .default) { _ in
            UIPasteboard.general.string = subChannel.id
        }
        
        AppUtility.showActionSheet(in: self, title: "", message: "Options", actions: [chatAction, editAction, viewDetail, subscribeAction, unSubscribeAction, copyAction])
    }
    
    func showEditScreen(subChannel: AmitySubChannel.ViewModel) {
        let editView = EditSubChannelView(viewModel: self.manager, subChannelId: subChannel.id, subChannelName: subChannel.displayName)
        let editHostingController = UIHostingController(rootView: editView)
        editHostingController.navigationController?.title = "Edit SubChannel"
        
        self.navigationController?.pushViewController(editHostingController, animated: true)
    }
    
    func displayChatScreen(subChannel: AmitySubChannel) {
        let storyboard = UIStoryboard(name: "Chats", bundle: nil)
        let viewController = storyboard.instantiateViewController(withIdentifier: "AmityChat")
        guard let chatViewController = viewController as? AmityChatViewController, let channel = self.channel else { return }
        
        chatViewController.channelId = channelId
        chatViewController.client = AmityManager.shared.client!
        chatViewController.subChannel = subChannel
        chatViewController.channelType = channel.channelType
        
        self.navigationController?.pushViewController(chatViewController, animated: true)
    }
}
